@@ -10,7 +10,7 @@ from logging import getLogger
 from os.path import isdir, isfile, join
 
 # create support
-from conda.base.constants import DepsModifier, UpdateModifier
+from conda.base.constants import ChannelPriority, DepsModifier, UpdateModifier
 from conda.base.context import context
 from conda.cli import common as cli_common
 from conda.cli.common import (
@@ -45,12 +45,14 @@ from conda.exceptions import (
 from conda.gateways.disk.create import mkdir_p
 from conda.gateways.disk.delete import delete_trash, path_is_clean, rm_rf
 from conda.gateways.disk.test import is_conda_environment
+from conda.lock import FileLock
 from conda.misc import explicit, touch_nonadmin
 from conda.models.match_spec import MatchSpec
 
 import mamba
 import mamba.mamba_api as api
 from mamba import repoquery as repoquery_api
+from mamba.mamba_shell_init import shell_init
 from mamba.utils import get_installed_jsonfile, init_api_context, load_channels, to_txn
 
 if sys.version_info < (3, 2):
@@ -139,13 +141,14 @@ def handle_txn(unlink_link_transaction, prefix, args, newenv, remove_op=False):
         raise DryRunExit()
 
     try:
-        unlink_link_transaction.download_and_extract()
-        if context.download_only:
-            raise CondaExitZero(
-                "Package caches prepared. UnlinkLinkTransaction cancelled with "
-                "--download-only option."
-            )
-        unlink_link_transaction.execute()
+        with FileLock(PackageCacheData.first_writable().pkgs_dir):
+            unlink_link_transaction.download_and_extract()
+            if context.download_only:
+                raise CondaExitZero(
+                    "Package caches prepared. UnlinkLinkTransaction cancelled with "
+                    "--download-only option."
+                )
+            unlink_link_transaction.execute()
 
     except SystemExit as e:
         raise CondaSystemExit("Exiting", e)
@@ -224,6 +227,9 @@ def remove(args, parser):
 
         solver_options.append((api.SOLVER_FLAG_ALLOW_UNINSTALL, 1))
 
+        if context.channel_priority is ChannelPriority.STRICT:
+            solver_options.append((api.SOLVER_FLAG_STRICT_REPO_PRIORITY, 1))
+
         pool = api.Pool()
         repos = []
 
@@ -239,18 +245,27 @@ def remove(args, parser):
             repos.append(repo)
 
         solver = api.Solver(pool, solver_options)
-        solver.add_jobs(mamba_solve_specs, api.SOLVER_ERASE)
+
+        history = api.History(context.target_prefix)
+        history_map = history.get_requested_specs_map()
+        solver.add_jobs(
+            [ms.conda_build_form() for ms in history_map.values()],
+            api.SOLVER_USERINSTALLED,
+        )
+
+        solver.add_jobs(mamba_solve_specs, api.SOLVER_ERASE | api.SOLVER_CLEANDEPS)
         success = solver.solve()
         if not success:
             print(solver.problems_to_str())
             exit_code = 1
             return exit_code
 
-        package_cache = api.MultiPackageCache(context.pkgs_dirs)
-        transaction = api.Transaction(
-            solver, package_cache, PackageCacheData.first_writable().pkgs_dir
-        )
-        downloaded = transaction.prompt(repos)
+        with FileLock(PackageCacheData.first_writable().pkgs_dir):
+            package_cache = api.MultiPackageCache(context.pkgs_dirs)
+            transaction = api.Transaction(
+                solver, package_cache, PackageCacheData.first_writable().pkgs_dir
+            )
+            downloaded = transaction.prompt(repos)
         if not downloaded:
             exit(0)
 
@@ -392,7 +407,7 @@ def install(args, parser, command="install"):
             )
             return
 
-        specs.extend([MatchSpec(s) for s in file_specs])
+        specs.extend(MatchSpec(s) for s in file_specs)
 
     specs.extend(specs_from_args(args_packages, json=context.json))
 
@@ -401,8 +416,8 @@ def install(args, parser, command="install"):
     for spec in specs:
         # CONDA TODO: correct handling for subdir isn't yet done
         spec_channel = spec.get_exact_value("channel")
-        if spec_channel and spec_channel not in channels:
-            channels.append(spec_channel)
+        if spec_channel and spec_channel.base_url not in channels:
+            channels.append(spec_channel.base_url)
 
     index_args["channel_urls"] = channels
 
@@ -420,8 +435,9 @@ def install(args, parser, command="install"):
     # and that they are name-only specs
     if isupdate and context.update_modifier == UpdateModifier.UPDATE_ALL:
         for i in installed_names:
-            if i != "python":
-                specs.append(MatchSpec(i))
+            if i.startswith("__"):
+                continue
+            specs.append(MatchSpec(i))
 
         prefix_data = PrefixData(prefix)
         for s in args_packages:
@@ -446,7 +462,7 @@ def install(args, parser, command="install"):
                 ms = MatchSpec(dep)
                 if ms.name != "python":
                     final_specs.append(MatchSpec(ms.name))
-        specs = set(final_specs)
+        specs = list(set(final_specs))
 
     if newenv and args.clone:
         if args.packages:
@@ -471,25 +487,28 @@ def install(args, parser, command="install"):
     if not (context.quiet or context.json):
         print("\nLooking for: {}\n".format([str(s) for s in specs]))
 
-    spec_names = [s.name for s in specs]
-
     # If python was not specified, check if it is installed.
     # If yes, add the installed python to the specs to prevent updating it.
     python_constraint = None
 
-    if "python" not in spec_names:
-        if "python" in installed_names:
-            i = installed_names.index("python")
-            version = installed_pkg_recs[i].version
-            python_constraint = MatchSpec("python==" + version).conda_build_form()
+    if "python" in installed_names:
+        if context.update_modifier == UpdateModifier.UPDATE_ALL or not any(
+            s.name == "python" for s in specs
+        ):
+            version = installed_pkg_recs[installed_names.index("python")].version
+            major_minor_version = ".".join(version.split(".")[:2])
+            python_constraint = f"python {major_minor_version}.*"
 
     mamba_solve_specs = [s.__str__() for s in specs]
 
+    if context.channel_priority is ChannelPriority.STRICT:
+        solver_options.append((api.SOLVER_FLAG_STRICT_REPO_PRIORITY, 1))
+
     pool = api.Pool()
 
     repos = []
 
-    if use_mamba_experimental or context.force_reinstall:
+    with FileLock(context.target_prefix):
         prefix_data = api.PrefixData(context.target_prefix)
         prefix_data.load()
 
@@ -526,6 +545,10 @@ def install(args, parser, command="install"):
                 (api.MAMBA_FORCE_REINSTALL, context.force_reinstall),
             ]
         )
+
+        if context.update_modifier is UpdateModifier.FREEZE_INSTALLED:
+            solver.add_jobs([p for p in prefix_data.package_records], api.SOLVER_LOCK)
+
         solver.add_jobs(mamba_solve_specs, solver_task)
 
         if not context.force_reinstall:
@@ -537,8 +560,8 @@ def install(args, parser, command="install"):
 
         pinned_specs_info = ""
         if python_constraint:
+            pinned_specs_info += f"  - {python_constraint}\n"
             solver.add_pin(python_constraint)
-            pinned_specs_info += f"  - {python_constraint}"
 
         pinned_specs = get_pinned_specs(context.target_prefix)
         if pinned_specs:
@@ -558,7 +581,7 @@ def install(args, parser, command="install"):
 
             try:
                 final_spec = s.conda_build_form()
-                pinned_specs_info += f"  - {final_spec}"
+                pinned_specs_info += f"  - {final_spec}\n"
                 solver.add_pin(final_spec)
             except AssertionError:
                 print(
@@ -567,7 +590,7 @@ def install(args, parser, command="install"):
                     "libname VERSION BUILD, for example libblas=*=*mkl\n"
                 )
 
-        if pinned_specs_info:
+        if pinned_specs_info and not (context.quiet or context.json):
             print(f"\nPinned packages:\n{pinned_specs_info}\n")
 
         success = solver.solve()
@@ -576,28 +599,29 @@ def install(args, parser, command="install"):
             exit_code = 1
             return exit_code
 
-        package_cache = api.MultiPackageCache(context.pkgs_dirs)
-        transaction = api.Transaction(
-            solver, package_cache, PackageCacheData.first_writable().pkgs_dir
-        )
-        mmb_specs, to_link, to_unlink = transaction.to_conda()
+        with FileLock(PackageCacheData.first_writable().pkgs_dir):
+            package_cache = api.MultiPackageCache(context.pkgs_dirs)
+            transaction = api.Transaction(
+                solver, package_cache, PackageCacheData.first_writable().pkgs_dir
+            )
+            mmb_specs, to_link, to_unlink = transaction.to_conda()
 
-        specs_to_add = [MatchSpec(m) for m in mmb_specs[0]]
-        specs_to_remove = [MatchSpec(m) for m in mmb_specs[1]]
+            specs_to_add = [MatchSpec(m) for m in mmb_specs[0]]
+            specs_to_remove = [MatchSpec(m) for m in mmb_specs[1]]
 
-        transaction.log_json()
+            transaction.log_json()
+            downloaded = transaction.prompt(repos)
+            if not downloaded:
+                exit(0)
 
-        downloaded = transaction.prompt(repos)
-        if not downloaded:
-            exit(0)
-        PackageCacheData.first_writable().reload()
+            PackageCacheData.first_writable().reload()
 
     # if use_mamba_experimental and not os.name == "nt":
     if use_mamba_experimental:
         if newenv and not isdir(context.target_prefix) and not context.dry_run:
             mkdir_p(prefix)
-
-        transaction.execute(prefix_data)
+        with FileLock(PackageCacheData.first_writable().pkgs_dir):
+            transaction.execute(prefix_data)
     else:
         conda_transaction = to_txn(
             specs_to_add,
@@ -704,6 +728,8 @@ def repoquery(args, parser):
         fmt = api.QueryFormat.JSON
     elif hasattr(args, "tree") and args.tree:
         fmt = api.QueryFormat.TREE
+    elif hasattr(args, "pretty") and args.pretty:
+        fmt = api.QueryFormat.PRETTY
     else:
         fmt = api.QueryFormat.TABLE
 
@@ -715,7 +741,6 @@ def repoquery(args, parser):
 
 def do_call(args, parser):
     relative_mod, func_name = args.func.rsplit(".", 1)
-
     # func_name should always be 'execute'
     if relative_mod in [
         ".main_list",
@@ -736,6 +761,8 @@ def do_call(args, parser):
         exit_code = create(args, parser)
     elif relative_mod == ".main_update":
         exit_code = update(args, parser)
+    elif relative_mod == ".main_init":
+        exit_code = shell_init(args)
     elif relative_mod == ".main_repoquery":
         exit_code = repoquery(args, parser)
     else:
@@ -776,6 +803,7 @@ Examples:
 
     package_cmds.add_argument("-p", "--platform")
     package_cmds.add_argument("--no-installed", action="store_true")
+    package_cmds.add_argument("--pretty", action="store_true")
 
     package_cmds.add_argument(
         "-a",
@@ -866,7 +894,8 @@ def main(*args, **kwargs):
     init_std_stream_encoding()
 
     if "activate" in sys.argv or "deactivate" in sys.argv:
-        print("Use conda to activate / deactivate the environment.")
+        print("Run 'mamba init' to be able to run mamba activate/deactivate")
+        print("and start a new shell session. Or use conda to activate/deactivate.")
         print("\n    $ conda " + " ".join(sys.argv[1:]) + "\n")
         return sys.exit(-1)
 
